
Object subclass: #Warehouse
    instanceVariableNames: 'products storages buffer '
    classVariableNames: ''
    poolDictionaries: ''!

Warehouse comment: 'Diese Klasse entspricht der Klasse Lager aus der OBA-Analyse.

Instanzvariablen:
	products <Set>
		Die im Lager vorhandenen ''Produkte''.
	storages <Dictionary>
		Die im Lager vorhandenen ''Regale''.
	buffer <WarehouseBuffer>
		Die ''Bereitstellungsflaeche'' des Lagers.
'!

!Warehouse class publicMethods !

example1
	"Ankunft einer Lieferung aus der Produktion.
	Warehouse example1"
	| warehouse supply store place |
	warehouse := Warehouse new.

	"Bereitstellflaeche wird eingerichtet."
	warehouse buffer initialize: 10 @ 4.

	"Regale werden installiert."
	1 to: 5 do:
		[:i | warehouse addStorage: (Storage new: 5 @ 4) inRow: i].

	"Naechste Lieferung wird erzeugt."
	supply := Set new.
	supply add: ((SupplyPalette new) product: (Product name: #'00001');
		suppliedQuantity: 100).
	supply add: ((SupplyPalette new) product: (Product name: #'00002');
		suppliedQuantity: 200).
	supply add: ((SupplyPalette new) product: (Product name: #'00003');
		suppliedQuantity: 300).
	supply add: ((SupplyPalette new) product: (Product name: #'00004');
		suppliedQuantity: 400).
	supply add: ((SupplyPalette new) product: (Product name: #'00005');
		suppliedQuantity: 500).
	supply add: ((SupplyPalette new) product: (Product name: #'00006');
		suppliedQuantity: 600).

	"Pruefe auf Bereitstellungsflaeche,
	 ob genug Stellflaeche vorhanden ist."
	(warehouse buffer hasSufficientSpaceFor: supply)
		ifFalse: [^self error: 'Not enough space to eccapt supply.'].

	"Lieferung wird entgegengenommen."
	supply do:
		[:palette |
			place := warehouse buffer choosePlace.
			warehouse buffer drop: palette at: place].

	"Zustand des Lagers wird mittels Inspektor gezeigt."
	warehouse inspect.

	"Lieferung wird in die Regale eingeordnet."
	supply do:
		[:palette |
			"Regal und Fach wird festgelegt und Palette abgelegt."
			store := warehouse chooseStorage.
			place := store choosePlace.
			(store at: place) store: palette.
			"Produkt und Menge auf Palette wird registriert."
			warehouse addProduct: palette product.
			palette product add: palette suppliedQuantity.
			"Platz auf Bereitstellungsflaeche wird wieder freigegeben."
			warehouse buffer clearFrom: palette].
	^warehouse!

example2
	"Bearbeiten eines Kundenauftrags.
	Warehouse example2"
	| warehouse order randomGenerator place currentUnit
		transUnitsToShip remainingQuantity shelf shelves |
	warehouse := Warehouse example1.
	randomGenerator := EsRandom new.
	order := CustomerOrder new.

	"Kunderauftrag wird zufaellig, eintsprechend der zur Verfuegung
	stehenden Produkte zusammengestellt."
	warehouse products do:
		[:product | |position|
			position := OrderPosition new.
			position product: product;
				quantity: (randomGenerator next * 100) rounded.
		order addPosition: position].

	"Lieferfaehigkeit wird ueberprueft."
	(warehouse canSatisfy: order)
		ifFalse: [^self error: 'Cannot deliver complete demand.'].

	"Kundenauftrag wird mittels Inspektor gezeigt."
	order inspect.

	"Transportbehaelter wird fuer den Kundenauftrag zur Verfuegung
	gestellt."
	currentUnit := TransportationUnit new.
	currentUnit customerOrder: order.

	"Platz auf der Bereistellungsflaeche wird gesucht."
	(place := warehouse buffer choosePlace) isNil
		ifTrue: [^self error: 'You have to wait for a place.'].

	"Transportbehaelter wird dort abgestellt."
	warehouse buffer drop: currentUnit at: place.

	"Transport wird zusammengestellt."
	transUnitsToShip := OrderedCollection with: currentUnit.

	"Jede einzelne Bestellung des Kundenauftrages wird bearbeitet."
	order orderPositions
		do: [:position |

			"Nachfrage nach dem Produkt wird verbucht."
			position product logOrderPosition: position.
			"Die Menge des Produkts der Bestellung wird ermittelt."
			remainingQuantity := position quantity.

			"Regale und Faecher, in denen das bestellten Produkt lagert,
			werden ermittelt."
			shelves := warehouse shelvesToServe: position.
			"Die Produktmenge wird entsprchend der Bestellung aus
			den Faechern herausgenommen und in Transportbehaelter
			abgelegt."
			shelves
				keysAndValuesDo:
					[:row :point | | shippingQuantity |

					"Das Fach aktuelle wird festgelegt."
					shelf := (warehouse storageInRow: row) at: point.
					"Die Menge des Produktes, die aus dem Fach
					genommen wird, wird bestimmt und entnommen."
					shippingQuantity := shelf storedQuantity
											min: remainingQuantity.
					shelf pick: shippingQuantity.
					"Ist die Palette leer, so wird sie aus dem Fach
					genommen und damit ist das Fach leer."
					shelf storedQuantity = 0 ifTrue: [shelf clear].

					"Entnommene Menge wird aus der Registratur
					des Lagers und von der Bestellung gestrichen."
					position product remove: shippingQuantity.
					remainingQuantity := remainingQuantity -
												shippingQuantity.

					"Das Produkt wird in den Transportbehaelter
					abgelegt und der noch zur Verfuegung stehende
					Platz im Behaelter bestimmt. Ist der Behaelter
					voll, so wird ein neuer zur Verfuegung gestellt."
					[shippingQuantity :=
						currentUnit pack: position product
									amount: shippingQuantity.
					shippingQuantity > 0]
						whileTrue:
							"Bereitstellen eines neuen Behaelters."
							[Transcript show: 'Need a new unit.'; cr.
							currentUnit := TransportationUnit new.
							"Registrierung eines neuen Behaelters
							zum Kundenauftrag."
							transUnitsToShip add: currentUnit.
							"Registrierung der zu ladenen Bestellung."
							currentUnit customerOrder: order.
							"Lagerung des Behaelters auf der
							Bereitstellungsflaeche der Lagers."
							(place := warehouse buffer choosePlace) isNil
								ifTrue:
								[^self error:'You have to wait for a place.'].
							warehouse buffer drop: currentUnit at: place]]].

	"Transportbehaelter sehen auf der Bereitstellungsflaeche des Lagers
	bereit und koennen mittels Inspektor ueberprueft werden."
	transUnitsToShip inspect.

	"Transportbehaelter werden versendet.
	Die Bereitstellungsflaeche wird wieder leer."
	transUnitsToShip
		do: [:trans |
			warehouse buffer ship: trans.
			warehouse buffer clearFrom: trans].
	^warehouse

!

new
	"Erzeugt ein Lager."
	^super new initialize
! !

Warehouse class categoriesFor: #example1 are:((OrderedCollection new)
	add: 'Examples';
	yourself)!
Warehouse class categoriesFor: #new are:((OrderedCollection new)
	add: 'Create';
	yourself)!
Warehouse class categoriesFor: #example2 are:((OrderedCollection new)
	add: 'Examples';
	yourself)!

!Warehouse publicMethods !

addProduct: aProduct
	"Hinzufuegen eines neuen Produkts."
	products add: aProduct
!

addStorage: aStorage inRow: aNumber
	"Hinzufuegen einer neuen Regalreihe."
	(storages at: aNumber ifAbsent: []) isNil
		ifFalse: [^self error: 'Row already in use.'].
	storages at: aNumber put: aStorage
!

buffer
	"Zugriff 'Bereitstellungsflaeche'."
	^buffer!

canSatisfy: aCustomerOrder
	"Entspricht der Dienstleistung 'Ausfuehrbar'."
	aCustomerOrder orderPositions do: [:position |
		position product availableQuantity < position quantity
			ifTrue: [^false]].
	^true!

chooseStorage
	"Entspricht der Dienstleistung 'Auswahl Regal'."
	"Es wird immer dasjenige Regal gewaehlt, das
	die geringste relative Belegung aufweist."
	| minUsage minStore |
	minUsage := 1.0.
	minStore := nil.
	storages do:
		[:store | 
			| newMin |
			(newMin := store usedSpace / store totalSpace) < minUsage
				ifTrue:
					[minUsage := newMin. minStore := store]].
	minStore isNil ifTrue:
		[^self error: 'No space to put item into storage.'].
	^minStore!

initialize
	"Anfangswerte werden gesetzt."
	products := EsIdentitySet new.
	storages := Dictionary new.
	buffer := WarehouseBuffer new: 1 @ 1!

products
	"Zugriff 'Produkte'."
	^products!

release
	"Objektdaten werden geloescht."
	products := nil.
	storages := nil.
	buffer := nil!

shelvesToServe: anOrderPosition
	"Entspricht der Dienstleistung 'Bedienfaecher'."
	"Die Methode liefert ein Dictionary mit Elementen
	der Form Regalnummer -> Fachnummer.
	Es mssen soviele Faecher bestimmt werden, dass deren
	summierte Inhalte mindestens die bestellte Menge erreichen.
	Dabei wird eine einfache Strategie verfolgt: Das erste Fach,
	das noch etwas von dem bestellten Produkt enthaelt, wird zur
	Bedienung des Bestellpostens herangezogen."
	| shelves remainingQuantity |
	shelves := Dictionary new.
	remainingQuantity := anOrderPosition quantity.
	(self products includes: anOrderPosition product)
		ifFalse: [^self error: 'Product not im marehouse.'].
	self storagesDo:
		[:row :store | store usedPlacesDo:
			[:point :shelf | shelf product == anOrderPosition product
				ifTrue:
					[remainingQuantity := remainingQuantity -
						shelf storedQuantity.
					shelves at: row put: point.
					remainingQuantity < 0 ifTrue: [^shelves]]]]!

storageInRow: aNumber
	"Zugriff 'Regal'."
	"Liefert das Regal mit der entsprechenden Nummer."
	| temporaries |
	^storages at: aNumber ifAbsent:[]!

storagesDo: aBinaryBlock
	"Wertet aBinaryBlock fuer jedes Regal aus.
	aBinaryBlock muss genau zwei Argumente haben. Das erste
	Argument ist die Reihe des Regals und das zweite
	Argument ist das Regal selbst."
	storages keysAndValuesDo:
		[:row :store | aBinaryBlock value: row value: store]! !

Warehouse categoriesFor: #addStorage:inRow: are:((OrderedCollection new)
	add: 'Accessing';
	yourself)!
Warehouse categoriesFor: #storageInRow: are:((OrderedCollection new)
	add: 'Accessing';
	yourself)!
Warehouse categoriesFor: #storagesDo: are:((OrderedCollection new)
	add: 'Enumerating';
	yourself)!
Warehouse categoriesFor: #initialize are:((OrderedCollection new)
	add: 'Initialize';
	yourself)!
Warehouse categoriesFor: #products are:((OrderedCollection new)
	add: 'Accessing';
	yourself)!
Warehouse categoriesFor: #shelvesToServe: are:((OrderedCollection new)
	add: 'Services';
	yourself)!
Warehouse categoriesFor: #buffer are:((OrderedCollection new)
	add: 'Accessing';
	yourself)!
Warehouse categoriesFor: #release are:((OrderedCollection new)
	add: 'Initialize';
	yourself)!
Warehouse categoriesFor: #canSatisfy: are:((OrderedCollection new)
	add: 'Services';
	yourself)!
Warehouse categoriesFor: #chooseStorage are:((OrderedCollection new)
	add: 'Services';
	yourself)!
Warehouse categoriesFor: #addProduct: are:((OrderedCollection new)
	add: 'Accessing';
	yourself)!